;*---------------------------------------------------------------------*/
;*    A practical implementation for the Scheme programming language   */
;*                                                                     */
;*                                    ,--^,                            */
;*                              _ ___/ /|/                             */
;*                          ,;'( )__, ) '                              */
;*                         ;;  //   L__.                               */
;*                         '   \\   /  '                               */
;*                              ^   ^                                  */
;*                                                                     */
;*               Copyright (c) 1992-1999 Manuel Serrano                */
;*                                                                     */
;*     Bug descriptions, use reports, comments or suggestions are      */
;*     welcome. Send them to                                           */
;*       bigloo@kaolin.unice.fr                                        */
;*       http://kaolin.unice.fr/bigloo                                 */
;*                                                                     */
;*   This program is free software; you can redistribute it            */
;*   and/or modify it under the terms of the GNU General Public        */
;*   License as published by the Free Software Foundation; either      */
;*   version 2 of the License, or (at your option) any later version.  */
;*                                                                     */
;*   This program is distributed in the hope that it will be useful,   */
;*   but WITHOUT ANY WARRANTY; without even the implied warranty of    */
;*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     */
;*   GNU General Public License for more details.                      */
;*                                                                     */
;*   You should have received a copy of the GNU General Public         */
;*   License along with this program; if not, write to the Free        */
;*   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,   */
;*   MA 02111-1307, USA.                                               */
;*---------------------------------------------------------------------*/
;*=====================================================================*/
;*    serrano/prgm/project/bdk/kbdb/kbdb/Bee/etags.scm                 */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Jul 29 09:18:33 1999                          */
;*    Last change :  Wed Aug 18 16:57:10 1999 (serrano)                */
;*    -------------------------------------------------------------    */
;*    The debugger tries to load the etags file for the project in     */
;*    order to be able to do some eager demangling. This is important  */
;*    in order to be able to set breakpoints using Bigloo symbolic     */
;*    names before the execution starts.                               */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module bee_etags
   (import engine_param
	   tools_speek
	   tools_error
	   (console-prompt tools_io)
	   (bdb-repl-prompt engine_repl))
   (export (read-etags-file)
	   (bigloo-symbol-etags-source-location ::bstring)))

;*---------------------------------------------------------------------*/
;*    scm-info                                                         */
;*---------------------------------------------------------------------*/
(define-struct scm-info
   ;; scm-name
   name
   ;; source file
   file
   ;; line number
   line)

;*---------------------------------------------------------------------*/
;*    bigloo-symbol-etags-source-location ...                          */
;*    -------------------------------------------------------------    */
;*    If we have found source file information for a Bigloo symbols    */
;*    reading the etags file, return that information.                 */
;*---------------------------------------------------------------------*/
(define (bigloo-symbol-etags-source-location scm-name::bstring)
   (if (hash-table? *etags-table*)
       (let ((scm (get-hash scm-name *etags-table*)))
	  (if (scm-info? scm)
	      (string-append (scm-info-file scm)
			     ":"
			     (integer->string (scm-info-line scm)))
	      #f))
       #f))

;*---------------------------------------------------------------------*/
;*    *etags-table* ...                                                */
;*---------------------------------------------------------------------*/
(define *etags-table* #unspecified)
   
;*---------------------------------------------------------------------*/
;*    read-etags-file ...                                              */
;*---------------------------------------------------------------------*/
(define (read-etags-file)
   (let ((file (find-file/path *etags* (list "." *root-directory*))))
      (if (and (string? file) (file-exists? file))
	  ;; we really have to read the etags file
	  (begin
	     (verbose 4 "reading etags file \"" file "\"...")
	     (set! *etags-table*
		   (make-hash-table 1024
				    (lambda (o) (string->0..2^x-1 o 10))
				    scm-info-name
				    string=?
				    64))
	     (read-etags file)
	     (verbose 4 #"done.\n")))))

;*---------------------------------------------------------------------*/
;*    read-etags ...                                                   */
;*    -------------------------------------------------------------    */
;*    An etags file format is:                                         */
;*    ^L                                                               */
;*    <file-name>,<etags-len>                                          */
;*    <definitions>+                                                   */
;*    <definition> : <keyword> <ident>^?<line-start>,<line-stop>       */
;*    <keyword>    : (define                                           */
;*                 | (define-generic                                   */
;*                 | (define-method                                    */
;*                 | (class                                            */
;*---------------------------------------------------------------------*/
(define (read-etags file-name)
   (if (not (file-exists? file-name))
       (bdb-error "etags"
		  "Can't find etags file for input"
		  file-name)
       (let ((port (open-input-file file-name)))
          (if (not (input-port? port))
              (bdb-error "etags"
			 "Can't find etags file for input"
			 file-name)
              (unwind-protect
                 (begin
                    ;; we skip the ^L line
                    (read-line port)
                    ;; then we read all file entries
                    (let loop ((entry (read-etags-file-entry port)))
		       (if (not (eof-object? entry))
			   (loop (read-etags-file-entry port)))))
                 (close-input-port port))))))

;*---------------------------------------------------------------------*/
;*    read-etags-file-entry ...                                        */
;*    -------------------------------------------------------------    */
;*    Because etags file format are just brain damaged, we rely on     */
;*    the assumption that reading at the end-of-file returns the       */
;*    eof-object. That is we read twice the end of file. In            */
;*    consequence it is impossible to use that function to read        */
;*    in stdin.                                                        */
;*---------------------------------------------------------------------*/
(define (read-etags-file-entry port)
   (define (fetch-file-name str)
      ;; fetch the file name of an etags file-name entry
      (let ((port  (open-input-string str))
            (fetch (regular-grammar ()
                      ((+ (out #\,))
                       (the-string))
                      (else
                       (bdb-error "etags"
				  "Illegal etags file format"
				  str)))))
         (let ((name (read/rp fetch port)))
            (close-input-port port)
            name)))
   ;; we read the file name
   (let ((line (read-line port)))
      (if (eof-object? line)
          line
          (let ((file (fetch-file-name line)))
	     (let loop ((line (read-line port)))
		(if (or (eof-object? line) (string=? line ""))
		    line
		    (begin
		       (match-case (parse-etags-entry-line line)
			  ((define (?name ?line))
			   ;; a function definition
			   (put-hash! (scm-info (string-upcase! name)
						file
						line)
				      *etags-table*))
			  ((define ?name ?line)
			   ;; a variable definition
			   (put-hash! (scm-info (string-upcase! name)
						file
						line)
				      *etags-table*))
			  ((define-generic (?name ?line))
			   ;; a generic function definition
			   (put-hash! (scm-info (string-upcase! name)
						file
						line)
				      *etags-table*))
			  ((extern ?name ?line)
			   ;; an extern definition
			   (put-hash! (scm-info (string-upcase! name)
						file
						line)
				      *etags-table*)))
		       (loop (read-line port)))))))))

;*---------------------------------------------------------------------*/
;*    parse-etags-entry-line ...                                       */
;*    -------------------------------------------------------------    */
;*    I don't know if it is the best way to proceed but there is a     */
;*    strong separation between the parsing of the etags file and      */
;*    the insertion in the hash-table. That is the function            */
;*    PARSE-ETAGS-ENTRY-LINE decodes the etags file and put it in      */
;*    some kind of intermediate representation. That representation    */
;*    is scan afterward and during that late scan, the insertions      */
;*    are operated if necessary.                                       */
;*---------------------------------------------------------------------*/
(define (parse-etags-entry-line line)
   (let ((port  (open-input-string line))
	 (reg   (regular-grammar ((letter   (in ("azAZ") (#a128 #a255)))
				  (special  (in "!@~$%^&*></-_+\\|=?.:"))
				  (kspecial (in "!@~$%^&*></-_+\\|=?."))
				  (id       (: (* digit)
					       (or letter special)
					       (* (or letter
						      special
						      digit
						      (in ",'`"))))))
		   (blank
		    (ignore))
		   ("("
		    (list 'PAR-OPEN))
		   (")"
		    (list 'PAR-CLO))
		   ("["
		    (list 'ANGLE-OPEN))
		   ("]"
		    (list 'ANGLE-CLO))
		   ("#"
		    (list 'SHARP))
		   (";"
		    (list 'SEMI-COLON))
		   ((+ digit)
		    (cons 'NUMBER (the-fixnum)))
		   (id
		    (let* ((string  (the-string))
			   (ustring (string-upcase string))
			   (symbol  (string->symbol string)) 
			   (kwd     (getprop symbol 'reserved)))
		       (if kwd
			   ;; this is a keyword
			   (cons kwd (string->symbol ustring))
			   ;; this is a regular identifier
			   (cons 'IDENT string))))
		   ((: #\" (* (out #\")) #\")
		    (list 'STRING))
		   (#a127
		    (list 'EOI))
		   (#\,
		    (ignore))
		   (else
		    (let ((c (the-failure)))
		       (if (eof-object? c)
			   c
			   (bdb-error "etags" "Illegal char" (char->integer c)))))))
	 (lalr  (lalr-grammar
		   (IDENT PAR-OPEN PAR-CLO ANGLE-OPEN ANGLE-CLO EOI NUMBER
			  DEFINE INLINE DEFMACRO
			  GENERIC METHOD
			  TYPESEP SHARP SEMI-COLON
			  CLASS STRUCT EXTERN MACRO STRING INCLUDE TYPE
			  STATIC/EXPORT)
		   
		   ;; the line we are to parse
		   (line
		    ((function)
		     function)
		    ((variable)
		     variable)
		    ((genericdef)
		     genericdef)
		    ((methoddef)
		     methoddef)
		    ((classdef)
		     classdef)
		    ((structdef)
		     structdef)
		    ((externdef)
		     externdef)
		    ((macrodef)
		     macrodef)
		    ((ignoredef)
		     '()))
		   
		   ;; function definitions
		   (function
		    ((PAR-OPEN DEFINE PAR-OPEN IDENT EOI NUMBER@line NUMBER)
		     `(define (,IDENT ,line)))
		    ((PAR-OPEN INLINE PAR-OPEN IDENT EOI NUMBER@line NUMBER)
		     `(define (,IDENT ,line))))
		   
		   ;; variable definitions
		   (variable
		    ((PAR-OPEN DEFINE IDENT EOI NUMBER@line NUMBER)
		     `(define ,IDENT ,line)))
		   
		   ;; generic function definitions
		   (genericdef
		    ((PAR-OPEN GENERIC PAR-OPEN IDENT EOI NUMBER@line NUMBER)
		     `(define-generic (,IDENT ,line))))
		   
		   ;; method definitions
		   (methoddef
		    ((PAR-OPEN METHOD PAR-OPEN IDENT@mnane IDENT@aname dummys
			       EOI NUMBER@line NUMBER)
		     `(define-method (,mnane ,aname ,line))))
		   
		   (dummys
		    (()
		     'dummy)
		    ((PAR-CLO dummys)
		     'dummy)
		    ((PAR-OPEN dummys)
		     'dummy)
		    ((IDENT dummys)
		     'dummy))
		   
		   ;; class definitions
		   (classdef
		    ((PAR-OPEN CLASS IDENT dummys
			       EOI NUMBER@line NUMBER)
		     `(,CLASS ,IDENT ,line))
		    ((PAR-OPEN STATIC/EXPORT
			       PAR-OPEN CLASS IDENT dummys
			       EOI NUMBER@line NUMBER)
		     `(,CLASS ,IDENT ,line)))
		   
		   ;; struct definitions
		   (structdef
		    ((PAR-OPEN STRUCT IDENT EOI NUMBER@line NUMBER)
		     `(define-struct ,IDENT ,line)))
		   
		   ;; extern definitions
		   (externdef
		    ((PAR-OPEN EXTERN extern-sans-clause)
		     extern-sans-clause)
		    ((extern-sans-clause)
		     extern-sans-clause))
		   
		   (extern-sans-clause
		    ((PAR-OPEN MACRO IDENT EOI NUMBER@line NUMBER)
		     `(extern ,IDENT ,line))
		    ((PAR-OPEN INCLUDE EOI NUMBER@line NUMBER)
		     `(ignore))
		    ((PAR-OPEN IDENT STRING EOI NUMBER@line NUMBER)
		     `(extern ,IDENT ,line))
		    ((PAR-OPEN IDENT EOI NUMBER@line NUMBER)
		     `(extern ,IDENT ,line))
		    ((PAR-OPEN STATIC/EXPORT IDENT STRING
			       EOI NUMBER@line NUMBER)
		     '(ignore))
		    ((PAR-OPEN TYPE EOI NUMBER NUMBER)
		     '(ignore)))
		   
		   ;; macro definitions
		   (macrodef
		    ((PAR-OPEN DEFMACRO PAR-OPEN IDENT EOI NUMBER@line NUMBER)
		     `(define-macro (,IDENT ,line))))

		   ;; ignored C definitions
		   (ignoredef
		    ((STATIC/EXPORT c-ignoredef)
		     'ignore)
		    ((c-ignoredef)
		     'ignore))

		   (c-ignoredef
		    ((IDENT IDENT SEMI-COLON EOI NUMBER@line NUMBER)
		     'ignore)
		    ((IDENT IDENT EOI NUMBER@line NUMBER)
		     'ignore)
		    ((SHARP DEFINE IDENT EOI NUMBER@line NUMBER)
		     'ignore)
		    ((IDENT IDENT PAR-OPEN EOI NUMBER@line NUMBER)
		     'ignore)
		    ((IDENT PAR-OPEN EOI NUMBER@line NUMBER)
		     'ignore)
	 	    ((IDENT ANGLE-OPEN EOI NUMBER@line NUMBER)
		     'ignore)
		    ((IDENT IDENT ANGLE-OPEN EOI NUMBER@line NUMBER)
		     'ignore)))))
		   
      (try (read/lalrp lalr reg port)
	   (lambda (escape obj proc msg)
	      (bdb-notify-error "etags" proc msg)
	      (console-prompt (bdb-repl-prompt))
	      (escape #f)))))

;*---------------------------------------------------------------------*/
;*    *keyword-list*                                                   */
;*---------------------------------------------------------------------*/
(define *keyword-list*
   '(("define" . DEFINE)
     ("define-inline" . INLINE)
     ("define-method" . METHOD)
     ("define-generic" . GENERIC)
     ("define-macro" . DEFMACRO)
     ("class" . CLASS) ("final-class" . CLASS) ("wide-class" . CLASS)
     ("extern" . EXTERN) ("macro" . MACRO)
     ("define-struct" . STRUCT)
     ("static" . STATIC/EXPORT) ("export" . STATIC/EXPORT)
     ("include" . INCLUDE)
     ("type" . TYPE)))

;*---------------------------------------------------------------------*/
;*    The keyword initialization                                       */
;*---------------------------------------------------------------------*/
(for-each (lambda (word)
	     (putprop! (string->symbol (car word)) 'reserved (cdr word)))
	  *keyword-list*)
