;*---------------------------------------------------------------------*/
;*    A practical implementation for the Scheme programming language   */
;*                                                                     */
;*                                    ,--^,                            */
;*                              _ ___/ /|/                             */
;*                          ,;'( )__, ) '                              */
;*                         ;;  //   L__.                               */
;*                         '   \\   /  '                               */
;*                              ^   ^                                  */
;*                                                                     */
;*               Copyright (c) 1992-1999 Manuel Serrano                */
;*                                                                     */
;*     Bug descriptions, use reports, comments or suggestions are      */
;*     welcome. Send them to                                           */
;*       bigloo@kaolin.unice.fr                                        */
;*       http://kaolin.unice.fr/bigloo                                 */
;*                                                                     */
;*   This program is free software; you can redistribute it            */
;*   and/or modify it under the terms of the GNU General Public        */
;*   License as published by the Free Software Foundation; either      */
;*   version 2 of the License, or (at your option) any later version.  */
;*                                                                     */
;*   This program is distributed in the hope that it will be useful,   */
;*   but WITHOUT ANY WARRANTY; without even the implied warranty of    */
;*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     */
;*   GNU General Public License for more details.                      */
;*                                                                     */
;*   You should have received a copy of the GNU General Public         */
;*   License along with this program; if not, write to the Free        */
;*   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,   */
;*   MA 02111-1307, USA.                                               */
;*---------------------------------------------------------------------*/
;*=====================================================================*/
;*    serrano/prgm/project/bdk/kbdb/src/Gdb/proc.scm                   */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Tue Jul 27 09:34:22 1999                          */
;*    Last change :  Tue Aug 31 09:08:07 1999 (serrano)                */
;*    -------------------------------------------------------------    */
;*    Starting and stopping the embedded GDB process.                  */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    The module                                                       */
;*---------------------------------------------------------------------*/
(module gdb_proc
   (include "Engine/hooks.sch")
   (import  engine_param
	    command_file
	    tools_error
	    tools_speek
	    tools_io
	    gdb_invoke
	    patient_mangling)
   (extern  (isatty?::bool    (::int)          "isatty")
	    (ttyname::string  (::int)          "ttyname")
	    (tty-init::string ()               "tty_init")
	    (tty-read::string ()               "tty_read")
	    (tty-write::int   (::string ::int) "tty_write"))
   (export  (gdb-start)
	    (gdb-stop)
	    (get-gdb-process)
	    (read-process-tty-output)
	    (write-process-tty-output str::bstring)))

;*---------------------------------------------------------------------*/
;*    *gdb-process* ...                                                */
;*---------------------------------------------------------------------*/
(define *gdb-process* #unspecified)

;*---------------------------------------------------------------------*/
;*    get-gdb-process ...                                              */
;*---------------------------------------------------------------------*/
(define (get-gdb-process)
   *gdb-process*)

;*---------------------------------------------------------------------*/
;*    gdb-start ...                                                    */
;*    -------------------------------------------------------------    */
;*    Start a new gdb process.                                         */
;*---------------------------------------------------------------------*/
(define (gdb-start)
   (if (> *verbose* 3)
       (verbose 4 "spawning : "
		(let ((port (open-output-string)))
		   (display *gdb* port)
		   (display " " port)
		   (for-each (lambda (x)
				(display x port)
				(display " " port))
			     (gdb-options-list))
		   (close-output-port port))
		#\Newline))
   (set! *gdb-process* (apply run-process
			      *gdb*
			      output: pipe:
			      input: pipe:
			      error: pipe:
			      (gdb-options-list)))
   (if (or (not (process? *gdb-process*))
	   (not (process-alive? *gdb-process*)))
       (bdb-error "gdb-start" "Can't start process" *gdb*)
       (begin
	  ;; initial gdb hook
	  (run-hooks *gdb-run-hook* *gdb-process*)
	  ;; flush
	  (flush-gdb-pending-output #t)
	  ;; we silently change the gdb configuration
	  (gdb-call-for-output "set print repeat -1"))))

;*---------------------------------------------------------------------*/
;*    gdb-stop ...                                                     */
;*---------------------------------------------------------------------*/
(define (gdb-stop)
   (if (and (process? *gdb-process*) (process-alive? *gdb-process*))
       (process-kill *gdb-process*)))

;*---------------------------------------------------------------------*/
;*    *tty-process?* ...                                               */
;*    -------------------------------------------------------------    */
;*    Is the running process using our own tty ?                       */
;*---------------------------------------------------------------------*/
(define *tty-process?* #f)

;*---------------------------------------------------------------------*/
;*    gdb-options-list ...                                             */
;*    -------------------------------------------------------------    */
;*    Build the standard gdb argument list. That list is the standard  */
;*    *GDB-OPTIONS* plus the information about the tty to be used by   */
;*    the process.                                                     */
;*---------------------------------------------------------------------*/
(define (gdb-options-list)
   (define (ttyopt)
      (let ((tname (and *force-tty?*
			(let ((name (tty-init)))
			   (if (run-hooks *tty-init-hook*)
			       name
			       #f)))))
	 (if (and (string? tname) (>fx (string-length tname) 0))
	     ;; we have successfully opened a tty
	     (begin
		(set! *tty-process?* #t)
		(string-append "--tty=" tname))
	     ;; we have been unable to open a tty
	     (begin
		(if (isatty? 0)
		    (let ((tty (ttyname 0)))
		       (string-append "--tty=" tty))
		    (error "gdb-options-list"
			   "Can't find any tty."
			   "Exiting."))))))
   (let ((opts (cons* (ttyopt) *gdb-options*)))
      (if (string? *exec*)
	  (begin
	     (reset-demangling-cache!)
	     (set-previous-file! *exec*)
	     (cons *exec* opts))
	  opts)))

;*---------------------------------------------------------------------*/
;*    read-process-tty-output ...                                      */
;*---------------------------------------------------------------------*/
(define (read-process-tty-output)
   (if (not *tty-process?*)
       #t
       (let ((str (tty-read)))
	  (if (>fx (string-length str) 0)
	      str
	      #f))))

;*---------------------------------------------------------------------*/
;*    write-process-tty-output ...                                     */
;*---------------------------------------------------------------------*/
(define (write-process-tty-output str::bstring)
   (if (not *tty-process?*)
       (error "write-process-tty-output"
	      "Can't write process input value"
	      str)
       (let ((len (string-length str)))
	  (tty-write str len))))
